@@ -28,10 +28,27 @@ class Connector:
     @staticmethod
     def connect(schema_name: str):
         logger.start(object={"schema_name": schema_name})
-        if schema_name in connections_pool and connections_pool[schema_name].connection.is_connected():
-            logger.end("Return existing connections_pool", object={
-                'connections_pool': str(connections_pool[schema_name])})
-            return connections_pool[schema_name]
+        if (schema_name in connections_pool and
+                connections_pool[schema_name] and
+                connections_pool[schema_name].connection):
+            if connections_pool[schema_name].connection.is_connected():
+                logger.end("Return existing connections_pool", object={
+                    'connections_pool': str(connections_pool[schema_name])})
+                return connections_pool[schema_name]
+            else:
+                # reconnect
+                connections_pool[schema_name].reconnect()
+                if connections_pool[schema_name].connection.is_connected():
+                    logger.end("Reconnected successfully", object={
+                        'connections_pool': str(connections_pool[schema_name])})
+                    return connections_pool[schema_name]
+                else:
+                    connector = Connector(schema_name)
+                    connections_pool[schema_name] = connector
+                    logger.error("Reconnect failed, returning a new connection",
+                                 object={'connections_pool': str(connections_pool[schema_name])})
+
+                    return connector._connect_to_db()
         else:
             connector = Connector(schema_name)
             connections_pool[schema_name] = connector
@@ -39,10 +56,13 @@ class Connector:
                        object={'connector': str(connector)})
             return connector._connect_to_db()
 
-    def __init__(self, schema_name, host=None, user=None, password=None) -> None: # noqa E501
-        if not all([os.getenv("RDS_HOSTNAME"), os.getenv("RDS_USERNAME"), os.getenv("RDS_PASSWORD")]): # noqa E501
+    def reconnect(self):
+        self.connection.reconnect()
+
+    def __init__(self, schema_name, host=None, user=None, password=None) -> None:
+        if not all([os.getenv("RDS_HOSTNAME"), os.getenv("RDS_USERNAME"), os.getenv("RDS_PASSWORD")]):
             logger.error(
-                "Error: Add RDS_HOSTNAME, RDS_USERNAME and RDS_PASSWORD to .env") # noqa E501
+                "Error: Add RDS_HOSTNAME, RDS_USERNAME and RDS_PASSWORD to .env")
             raise Exception("Environment variables not set.")
 
         self.host = host or os.getenv("RDS_HOSTNAME")
@@ -51,9 +71,9 @@ class Connector:
         self.password = password or os.getenv("RDS_PASSWORD")
 
         # Checking RDS_HOSTNAME suffix
-        if not (self.host.endswith("circ.zone") or self.host.endswith("circlez.ai")): # noqa E501
+        if not (self.host.endswith("circ.zone") or self.host.endswith("circlez.ai")):
             logger.error(
-                f"Warning: Your RDS_HOSTNAME={self.host} which is not what is expected") # noqa E501
+                f"Warning: Your RDS_HOSTNAME={self.host} which is not what is expected")
         self.connection: mysql.connector = None
         self._cursor = None
         logger.end()
@@ -125,7 +145,7 @@ class Connector:
                 raise
         else:
             logger.error(
-                "Connection is not established. The database will be used on the next connect.") # noqa E501
+                "Connection is not established. The database will be used on the next connect.")
         logger.end()
 
     def rollback(self):
@@ -136,4 +156,4 @@ class Connector:
     def start_tranaction(self):
         logger.start()
         self.connection.start_transaction()
-        logger.end()
\ No newline at end of file
+        logger.end()
