@@ -6,7 +6,7 @@ from logger_local.LoggerComponentEnum import LoggerComponentEnum    # noqa E402
 
 
 DATABASE_MYSQL_PYTHON_PACKAGE_COMPONENT_ID = 13
-DATABASE_MYSQL_PYTHON_PACKAGE_COMPONENT_NAME = 'circles_local_database_python'    # noqa E501
+DATABASE_MYSQL_PYTHON_PACKAGE_COMPONENT_NAME = 'circles_local_database_python'
 DEVELOPER_EMAIL = 'valeria.e@circ.zone and idan.a@circ.zone'
 obj = {
     'component_id': DATABASE_MYSQL_PYTHON_PACKAGE_COMPONENT_ID,
@@ -22,9 +22,9 @@ class Cursor:
     def __init__(self, cursor) -> None:
         self.cursor = cursor
 
-    # TODO: If environment <> prod1 and dvlp1 break down using 3rd party package and analyze the formatted_sql  # noqa E501
+    # TODO: If environment <> prod1 and dvlp1 break down using 3rd party package and analyze the formatted_sql
     #  and call private method _validate_select_table_name(table_name)
-    def execute(self, sql_statement: str, sql_parameters: tuple = None) -> None:    # noqa E501
+    def execute(self, sql_statement: str, sql_parameters: tuple = None) -> None:
         # TODO: validate_select_table_name(table_name)
         object1 = {
             "sql_statement": sql_statement,
@@ -39,7 +39,7 @@ class Cursor:
         else:
             formatted_sql = sql_statement
             sql_parameters_str = "None"
-        EXECUTE_METHOD_NAME = 'database-mysql-local-python-package cursor.py execute()'   # noqa E501
+        EXECUTE_METHOD_NAME = 'database-mysql-local-python-package cursor.py execute()'
         logger.info(EXECUTE_METHOD_NAME, object={
             "full_sql_query": formatted_sql,
             "sql_parameters": sql_parameters_str,
